/* giFTui
 * Copyright (C) 2003 the giFTui team
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "main.h"

#include <string.h>
#include <stdarg.h>
#include <glib-object.h>

#include "util.h"
#include "signal.h"

#define GIFTUI_SIGNAL(arg) ((GiftuiSignal_t *) (arg))
#define GIFTUI_HANDLER(arg) ((GiftuiHandler_t *) (arg))

typedef struct _GiftuiHandler_t
{
	GiftuiHandlerCB_t handler;
	gpointer data;
} GiftuiHandler_t;

typedef struct _GiftuiSignal_t
{
	gchar *name;
	GList *handlers;
} GiftuiSignal_t;

/**/

static GList *signals_list;

/* Handlers */

static GiftuiHandler_t *
giftui_handler_new (GiftuiHandlerCB_t handler, gpointer data)
{
	GiftuiHandler_t *hand;
	
	hand = g_new0 (GiftuiHandler_t, 1);
	g_return_val_if_fail (hand != NULL, hand);
	
	hand->handler = handler;
	hand->data = data;
	
	return hand;
}

static void
giftui_handler_free (GiftuiHandler_t *hand)
{
	g_return_if_fail (hand != NULL);
	
	g_free (hand);
	
	return;
}

/**/

static gint
_handler_find_by_handler (GiftuiHandler_t *hand1, GiftuiHandler_t *hand2)
{
	if (hand1->handler != hand2->handler)
		return TRUE;
	
	return FALSE;
}

static gint
_handler_find_by_data (GiftuiHandler_t *hand1, GiftuiHandler_t *hand2)
{
	if (hand1->data != hand2->data)
		return TRUE;
	
	return FALSE;
}

static gint
_handler_find_by_data_handler (GiftuiHandler_t *hand1, GiftuiHandler_t *hand2)
{
	return (_handler_find_by_handler (hand1, hand2) ||
		_handler_find_by_data (hand1, hand2));
}

/* Signals */

static GiftuiSignal_t *
giftui_signal_new (const gchar *name)
{
	GiftuiSignal_t *sig;
	
	g_return_val_if_fail (name != NULL, NULL);
	
	sig = g_new0 (GiftuiSignal_t, 1);
	g_return_val_if_fail (sig != NULL, NULL);
	
	sig->name = g_strdup (name);
	
	return sig;
}

static void
giftui_signal_free (GiftuiSignal_t *sig)
{
	g_return_if_fail (sig != NULL);
	
	g_list_foreach (g_list_first (sig->handlers),
			(GFunc) giftui_handler_free,
			NULL);
	g_list_free (g_list_first (sig->handlers));
	
	g_free (sig->name);
	g_free (sig);
	
	return;
}

static GiftuiSignal_t *
giftui_signal_add (const gchar *name)
{
	GiftuiSignal_t *sig;
	
	g_return_val_if_fail (name != NULL, NULL);
	
	if ((sig = giftui_signal_new (name)))
	{
		signals_list = g_list_append (signals_list, sig);
		signals_list = g_list_first (signals_list);
	}
	
	return sig;
}

static void
giftui_signal_remove (GList *el)
{
	g_return_if_fail (el != NULL);
	
	signals_list = g_list_remove_link (signals_list, el);
	signals_list = g_list_first (signals_list);
	
	giftui_signal_free (GIFTUI_SIGNAL (el->data));
	g_list_free_1 (el);
	
	return;
}

static void
giftui_signal_add_handler (GiftuiSignal_t *sig, GiftuiHandler_t *hand)
{
	g_return_if_fail (sig != NULL);
	g_return_if_fail (hand != NULL);
	
	sig->handlers = g_list_append (sig->handlers, hand);
	sig->handlers = g_list_first (sig->handlers);
	
	return;
}

static void
giftui_signal_remove_handler (GiftuiSignal_t *sig, GList *el)
{
	g_return_if_fail (sig != NULL);
	g_return_if_fail (el != NULL);
	
	sig->handlers = g_list_remove_link (sig->handlers, el);
	sig->handlers = g_list_first (sig->handlers);
	
	giftui_handler_free (GIFTUI_HANDLER (el->data));
	g_list_free_1 (el);
	
	return;
}

/**/

static gint
_signal_find_by_name (GiftuiSignal_t *sig, gchar *name)
{
	return strcmp (sig->name, name);
}

/* PUBLIC */

void
giftui_signal_register (const gchar *name, GiftuiHandlerCB_t handler, gpointer data)
{
	GList *l;
	GiftuiSignal_t *sig;
	GiftuiHandler_t *hand;
	
	g_return_if_fail (name != NULL);
	
	l = g_list_find_custom (g_list_first (signals_list),
				name,
				(GCompareFunc) _signal_find_by_name);
	
	if (l)
		sig = GIFTUI_SIGNAL (l->data);
	else
		sig = giftui_signal_add (name);
	
	if (sig && (hand = giftui_handler_new (handler, data)))
		giftui_signal_add_handler (sig, hand);
	
	return;
}

void
giftui_signal_unregister (const gchar *name, GiftuiHandlerCB_t handler, gpointer data)
{
	GList *e, *ls, *lh;
	GiftuiHandler_t sear;
	GiftuiSignal_t *sig;
	
	g_return_if_fail (name != NULL);
	
	ls = g_list_find_custom (g_list_first (signals_list),
				name,
				(GCompareFunc) _signal_find_by_name);
	
	/* Signal doesn't exist. */
	if (!ls)
		return;
	
	sig = GIFTUI_SIGNAL (ls->data);
	
	sear.handler = handler;
	sear.data = data;
	
	lh = g_list_first (sig->handlers);
	while (lh && (e = g_list_find_custom (lh,
					     &sear,
					     (GCompareFunc) _handler_find_by_data_handler)))
	{
		lh = e->next;
		giftui_signal_remove_handler (sig, e);
	}
	
	/* No more handler ? So we remove the signal. */
	if (!sig->handlers)
		giftui_signal_remove (ls);
	
	return;
}

void
giftui_signal_unregister_by_handler (const gchar *name, GiftuiHandlerCB_t handler)
{
	GList *e, *ls, *lh;
	GiftuiHandler_t sear;
	GiftuiSignal_t *sig;
	
	g_return_if_fail (name != NULL);
	
	ls = g_list_find_custom (g_list_first (signals_list),
				 name,
				 (GCompareFunc) _signal_find_by_name);
	
	/* Signal doesn't exist. */
	if (!ls)
		return;
	
	sig = GIFTUI_SIGNAL (ls->data);
	
	sear.handler = handler;
	
	lh = g_list_first (sig->handlers);
	while (lh && (e = g_list_find_custom (lh,
					     &sear,
					     (GCompareFunc) _handler_find_by_handler)))
	{
		lh = e->next;
		giftui_signal_remove_handler (sig, e);
	}
	
	/* No more handler ? So we remove the signal. */
	if (!sig->handlers)
		giftui_signal_remove (ls);
	
	return;
}

void
giftui_signal_unregister_by_data (const gchar *name, gpointer data)
{
	GList *e, *ls, *lh;
	GiftuiHandler_t sear;
	GiftuiSignal_t *sig;
	
	g_return_if_fail (name != NULL);
	
	ls = g_list_find_custom (g_list_first (signals_list),
				 name,
				 (GCompareFunc) _signal_find_by_name);
	
	/* Signal doesn't exist. */
	if (!ls)
		return;
	
	sig = GIFTUI_SIGNAL (ls->data);
	
	sear.data = data;
	
	lh = g_list_first (sig->handlers);
	while (lh && (e = g_list_find_custom (lh,
					     &sear,
					     (GCompareFunc) _handler_find_by_data)))
	{
		lh = e->next;
		giftui_signal_remove_handler (sig, e);
	}
	
	/* No more handler ? So we remove the signal. */
	if (!sig->handlers)
		giftui_signal_remove (ls);
	
	return;
}

/**/

void
giftui_signal_emit (const gchar *name)
{
       	GList *l;
	GiftuiHandler_t *hand;
	
	g_return_if_fail (name != NULL);
	
	l = g_list_find_custom (g_list_first (signals_list),
				name,
				(GCompareFunc) _signal_find_by_name);
	
	if (!l)
		return;
	
	l = GIFTUI_SIGNAL (l->data)->handlers;
	while (l)
	{
		hand = GIFTUI_HANDLER (l->data);
		if (hand && hand->handler)
			hand->handler (hand->data);
		l = l->next;
	}
	
	return;
}

/**/

void
giftui_signal_init (void)
{
	signals_list = NULL;
	
	return;
}

void
giftui_signal_end (void)
{
	g_list_foreach (g_list_first (signals_list),
			(GFunc) giftui_signal_free,
			NULL);
	
	return;
}
